#include <memory>
#include <list>
#include <iostream>

struct Vec3 {};
struct Quat {};
struct Mat44 {};

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

class Component;
class Frame;

using ComponentPtr = std::shared_ptr<Component>;
using ComponentWeakPtr = std::weak_ptr<Component>;
using FramePtr = std::shared_ptr<Frame>;

class Component {
    bool active = true;

    friend class GameNode;
    GameNode* node = nullptr;

public:
    virtual ~Component() {}

    bool is_active() const;
    void set_active(bool state); // Calls on(De)Activated()

    GameNode& get_node() const;

    virtual void initialize() {}
    virtual void release() {}

    virtual void onActivated() {}
    virtual void onDeactivated() {}
};

GameNode& Component::get_node() const {
    return *node;
}


class Frame : public Component { // Frame of reference
    FramePtr parent{};

    Vec3 origin{};
    Quat orientation{};

    Vec3 scale{}; // Optional

    mutable Mat44* to_world{ nullptr };
    mutable Mat44* from_world{ nullptr };

    void compute_to_world_matrix() const;
    void compute_from_world_matrix() const;
    void delete_matrices();

public:
    ~Frame();

    void set_parent(FramePtr frame, bool keep_transform);

    Mat44 const& get_to_world() const;
    Mat44 const& get_from_world() const;
};

Frame::~Frame() {
    delete_matrices();
}

void Frame::set_parent(FramePtr frame, bool keep_transform) {
    if (frame == parent)
        return;
    delete_matrices();
    // ...
}

Mat44 const& Frame::get_to_world() const {
    compute_to_world_matrix();
    return *to_world;
}

void Frame::compute_to_world_matrix() const {
    if (to_world != nullptr)
        return;
    if (parent != nullptr)
        parent->compute_to_world_matrix();
    to_world = new Mat44;
    // ...
}

void Frame::compute_from_world_matrix() const {
    if (from_world != nullptr)
        return;
    // ...
}

void Frame::delete_matrices() {
    delete to_world;
    delete from_world;
    to_world = nullptr;
    from_world = nullptr;
}


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

class GameNode;
class GameNodeHierarchy;

using GameNodePtr = std::shared_ptr<GameNode>;
using GameNodeWeakPtr = std::weak_ptr<GameNode>;

class GameNode {
    GameNodeWeakPtr parent{};
    std::list<GameNodePtr> children{};
    std::list<ComponentPtr> components{};

    // Only the GameNodeHierarchy can modify the tree.
    friend class GameNodeHierarchy;

    void push_back_child(GameNodePtr child);
    void erase_subtree(GameNodePtr child);

public:
    virtual ~GameNode() {}

    GameNodePtr get_parent() const;
    std::list<GameNodePtr> const& get_children() const;
    std::list<ComponentPtr> const& get_components() const;

    template<typename ComponentType>
    std::shared_ptr<ComponentType> find_component() const;

    void insert_frame(Vec3 origin, Quat orientation);
    void push_back_component(ComponentPtr component);
};

void GameNode::push_back_child(GameNodePtr child) {
    children.push_back(child);
}

template<typename ComponentType>
std::shared_ptr<ComponentType> GameNode::find_component() const {
    for (ComponentPtr component : components)
        if (auto ptr = std::dynamic_pointer_cast<ComponentType>(component))
            return ptr;
    return nullptr;
}

void GameNode::insert_frame(Vec3 origin, Quat orientation) {
    components.push_front(std::make_shared<Frame>());
    components.front()->node = this;
}

void GameNode::push_back_component(ComponentPtr component) {
    components.push_back(component);
    component->node = this;
}


class GameNodeHierarchy { // The only owner of all nodes!
    GameNodePtr root;

    // Ensuring only one instance exists.
    GameNodeHierarchy();
    GameNodeHierarchy(GameNodeHierarchy const&) = default;
    GameNodeHierarchy(GameNodeHierarchy&&) = default;
    GameNodeHierarchy& operator=(GameNodeHierarchy const&) = default;
    GameNodeHierarchy& operator=(GameNodeHierarchy&&) = default;

public:
    static GameNodeHierarchy& instance();

    GameNodePtr get_root() const;

    template<typename NodeType, class... ParameterTypes>
    std::shared_ptr<NodeType> push_back_child(GameNodePtr parent, ParameterTypes... args);
};

GameNodeHierarchy::GameNodeHierarchy() {
    root = std::make_shared<GameNode>();
}

GameNodeHierarchy& GameNodeHierarchy::instance() {
    static GameNodeHierarchy inst;
    return inst;
}

GameNodePtr GameNodeHierarchy::get_root() const {
    return root;
}

template<typename NodeType, class... ParameterTypes>
std::shared_ptr<NodeType> GameNodeHierarchy::push_back_child(GameNodePtr parent, ParameterTypes... args)
{
    auto child = std::make_shared<NodeType>(args...);
    parent->push_back_child(child);
    return child;
}

inline GameNodeHierarchy& hierarchy() { return GameNodeHierarchy::instance(); }

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

class SphereCollider : public Component {
    float radius;
public:
    SphereCollider(float radius_);
};

using SphereColliderPtr = std::shared_ptr<SphereCollider>;

SphereCollider::SphereCollider(float radius_)
    : radius{ radius_ }
{}

class Physics {

    // Ensuring only one instance exists.
    Physics() = default;
    Physics(Physics const&) = default;
    Physics(Physics&&) = default;
    Physics& operator=(Physics const&) = default;
    Physics& operator=(Physics&&) = default;

    std::list<SphereColliderPtr> sphereColliders;

public:
    static Physics& instance();

    SphereColliderPtr create_sphere_collider(float radius);

    void update();
};

Physics& Physics::instance() {
    static Physics phx;
    return phx;
}

SphereColliderPtr Physics::create_sphere_collider(float radius) {
    sphereColliders.push_back(std::make_shared<SphereCollider>(radius));
    return sphereColliders.back();
}

void Physics::update() {
    std::cout << "Physics: updating " << sphereColliders.size() << " spheres." << std::endl;
}

inline Physics& phx() { return Physics::instance(); }


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

class GameScript : public Component {
public:
    virtual void update() {}
};

using GameScriptPtr = std::shared_ptr<GameScript>;

class ScriptingEngine {

    // Ensuring only one instance exists.
    ScriptingEngine() = default;
    ScriptingEngine(ScriptingEngine const&) = default;
    ScriptingEngine(ScriptingEngine&&) = default;
    ScriptingEngine& operator=(ScriptingEngine const&) = default;
    ScriptingEngine& operator=(ScriptingEngine&&) = default;

    std::list<GameScriptPtr> scripts;

public:
    static ScriptingEngine& instance();

    template<typename ScriptType, class... ParameterTypes>
    std::shared_ptr<ScriptType> create_script(ParameterTypes... args);

    void update();
};

ScriptingEngine& ScriptingEngine::instance() {
    static ScriptingEngine phx;
    return phx;
}

template<typename ScriptType, class... ParameterTypes>
std::shared_ptr<ScriptType> ScriptingEngine::create_script(ParameterTypes... args) {
    auto script = std::make_shared<ScriptType>(args...);
    scripts.push_back(script);
    return script;
}

void ScriptingEngine::update() {
    std::cout << "ScriptingEngine: calling update method of " << scripts.size() << " scripts..." << std::endl;
    for (auto script_ptr : scripts)
        script_ptr->update();
}

inline ScriptingEngine& scripting() { return ScriptingEngine::instance(); }

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

class MySphereControllerScript : public GameScript {
public:
    virtual void update() {
        std::cout << "MySphereControllerScript: checking contents..." << std::endl;
        std::cout << "   frame found: " << (get_node().find_component<Frame>() != nullptr) << std::endl;
        std::cout << "   sphere collider found: " << (get_node().find_component<SphereCollider>() != nullptr) << std::endl;
    }
};

void initialize_game() {
    hierarchy().get_root()->insert_frame({}, {});
    hierarchy().get_root()->push_back_component(
        phx().create_sphere_collider(1.0f)
    );
    hierarchy().get_root()->push_back_component(
        scripting().create_script<MySphereControllerScript>()
    );
}

bool is_game_finished() {
    return true;
}


void release_game() {
    // ...
}

void  game_loop() {
    initialize_game();
    do {
        phx().update();
        scripting().update();
    } while (!is_game_finished());
    release_game();
}

int main()
{
    game_loop();
    return 0;
}
